{% extends "layout.html" %}

{% block document %}
<section class="hero is-bold is-medium">
    <div class="hero-body home-hero">
        <div class="container has-text-centered">
            <div class="columns is-vcentered">
                <div class="column is-5 is-offset-1">
                    <figure class="image is-4by3">
                        <img src="{{ pathto('_static/' + logo, 1) }}" class="logo" alt="TensorLy logo">
                    </figure>
                </div>

                <div class="column is-5 is-offset-1">
                    <h1 class="title is-1">
                        TensorLy
                    </h1>
                    <hr/>
                    <h2 class="subtitle is-3">
                        Simple and Fast Tensor Learning in Python
                    </h2>
                    <div class="container tensorly-functionalities-container">
                        <div class="content tensorly-functionalities">
                        <ul>
                            <li> Minimal dependencies </li>
                            <li> Tested and optimised </li>
                            <li> Thorough documentation </li>
                            <li> Pure Python </li>
                            <li> NumPy, PyTorch and MXnet backends </li>
                        </ul>
                        </div>
                    </div>
                    <p class="has-text-centered">
                    <a class="button is-large is-primary is-outlined" href="{{pathto('installation')}}">
                        Get started
                    </a>
                    </p>
                </div>
            </div>
        </div>
    </div>


    <div class="hero-cta">
        <div class="container has-text-centered">
            <p class="title">Discover TensorLy's functionalities!</p>
        </div>
        <div class="container has-text-centered">
            <br/>
            <a href="#functionalities">
            <span>
                <i class="fa fa-chevron-down" aria-hidden="true"></i>
            <span>
            </a>
        </div>
    </div>


    <div class="section home-main">
        <a name="functionalities" style="visibility: hidden;"></a>

        <div class="container">
            <div class="tile is-ancestor">
                    <div class="tile is-parent">
                    <a class="tile is-child box" href="{{pathto('installation')}}">
                        <p class="title">Install TensorLy</p>
                        <p class="subtitle">Installation instructions </p>
                    </a>
                    </div>

                    <div class="tile is-parent">
                    <a class="tile is-child box" href="{{pathto('user_guide/index')}}">
                        <p class="title">User Guide</p>
                        <p class="subtitle">A friendly guide to tensor learning</p>
                    </a>
                    </div>

                    <div class="tile is-parent">
                    <a class="tile is-child box" href= "{{pathto('auto_examples/index')}}">
                        <p class="title">Examples</p>
                        <p class="subtitle">See usage examples with code</p>
                    </a>
                    </div>
            </div>
            
            <div class="tile is-ancestor">
                    <div class="tile is-parent">
                    <a class="tile is-child box" href="{{pathto('modules/api')}}">
                        <p class="title">
                            <i class="fa fa-book" aria-hidden="true"></i>
                            API
                        </p>
                        <p class="subtitle">Functions and classes documentation</p>
                    </a>
                    </div>

                    <div class="tile is-parent">
                    <a class="tile is-child box" href="{{pathto('authors')}}">
                        <p class="title">Authors</p>
                        <p class="subtitle">About the developers</p>
                    </a>
                    </div>

                    <div class="tile is-parent">
                    <a class="tile is-child box" href= "https://github.com/tensorly/tensorly">
                        <p class="title">
                            <span class="icon"><i class="fa fa-github"></i></span>
                            <span>Contribute on Github</span>
                        </p>
                        <p class="subtitle">Source code</p>
                    </a>
                    </div>
            </div>
        </div>

    </div>
</section>



{% include 'footer.html' %}

{% endblock document %}

