<footer class="footer">
    <div class="container has-text-centered">

        <p>
        {%- if show_copyright %}
            {%- if hasdoc('copyright') %}
                {% trans path=pathto('copyright'), copyright=copyright|e %}&copy; <a href="{{ path }}">Copyright</a> {{ copyright }}.
                {% endtrans %}
            {%- else %}
                {% trans copyright=copyright|e %}&copy; {{ copyright }}.{% endtrans %}
            {%- endif %}
        {%- endif %}
        </p>

        {%- if show_sphinx %}
            {% trans %}Built with <a href="http://sphinx-doc.org/">Sphinx</a> using the TensorLy <a href="https://github.com/tensorly/tensorly">TensorLy</a> theme{% endtrans %}.
        {%- endif %}

        {%- block extrafooter %} {% endblock %}
    </div>
</footer>
