<nav class="pagination">
    {# Next and Previous #}
    {% if prev %}
    <a class="button is-medium pagination-previous" href="{{ prev.link|e }}" title="{{ prev.title|striptags|e }}" accesskey="p">
        <span class="icon">
            <i class="fa fa-arrow-circle-left"></i>
        </span>
        <span>{{ _('Previous') }}</span>
    </a>
    {% endif %}
    {% if next %}
    <a class="button is-medium pagination-next" href="{{ next.link|e }}" title="{{ next.title|striptags|e }}" accesskey="n">
        <span>{{ _('Next') }} </span>
        <span class="icon">
            <i class="fa fa-arrow-circle-right"></i>
        </span>
    </a>
    {% endif %}
</nav>

