/* *******
 * GENERAL
 * ******* */
body {
    font-family: -apple-system,BlinkMacSystemFont,"Segoe UI",Roboto,Helvetica,Arial,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol";
    overflow: hidden!important; /* prevents side scrolling past the limit on phones */
}
.footer {
    margin-bottom: 0;
}
.pagination-next, .pagination-previous {
    margin: 1em;
}
a {
    color: #03b198;
}
.main-column {
    padding: 0 0;
    background-color: whitesmoke;
    min-height: calc(100vh + 1rem);
}
/* API table */
td {
	display: table-cell!important;
}
th {
      background: whitesmoke;
  }
@media screen and (max-width: 768px) {
	table, tbody, td, tr {
		display: block!important; /* stack on mobile */
        overflow-wrap: break-word;
	}
	td {
      overflow-wrap: break-word;
  }
}

/* ****
 * MATH
 * **** */
div.math {
    text-align: center;
}


/* **********
 * TOP NAVBAR
 * ********** */
.navbar {
	margin-bottom: 0.2em; /* let shodow appear */
}
.navbar-item {j
    padding-bottom: 0.5em;
    font-size: 1.5em;
}
.navbar-title {
    font-size: 2.75rem;
}
.navbar-burger {
	margin-top: 1rem;
}
@media screen and (min-width: 1008px) { /* Leave some space on large screens */
	.navbar-end {
		padding-left: 1em;
	}
}


/* *******
 * CONTENT
 * ******* */
.content {
    display: block;
    background-color: #fff;
    padding: 1.5rem 1.5rem; /* we removed padding between sections... Compensating here */
}
.content .section {
	padding: 0; /* prevent each new subsection to be smaller (especially horrible on mobiles where the sub-sub-section becomes a tiny column */
}
.content h1 {
    padding-bottom: 0.4em;
    padding-top: 0.4em;
    font-size: 2.5rem;
	overflow-wrap: break-word;
}
.content h2 {
    padding-top: 0.6em;
	overflow-wrap: break-word;
}
.content h3 {
	padding-top: 0.5em;
	overflow-wrap: break-word;
}
.content blockquote {
    background-color: transparent;
    border-left: transparent;
    padding: 0 1em;
}
.main-content {
    position: relative;
    background-color: #fff;
}

/* ******************
 * NAVIGATION SIDEBAR
 * ****************** */

/* POSITIONING THE SIDEBAR */
.columns {
    margin: 0!important;
}
.aside {
    padding: inherit;
}
.sidebar {
    background-color: #232B2D;
    height: 100vh;
	overflow: auto; /* enable scrolling */
}
.stick {
    position: fixed;
    top: 0px;
	width: 25%;
}
.toc {
    padding-top: 0.5em;
    padding-bottom: 3em;
}
.search {
    padding-bottom: 1em;
    padding-top: 2em;
    padding-left: 2em;
    padding-right: 1em;
}
.search-title {
    font-weight: bold;
    font-size: 1.1em;
    color: lightgray;
}

/* STYLING THE TOC */
.show {
    display: block!important;
}
.sidebar .no-expand a:before {
	    content: '\00a0'!important; /* do not show expand icon */
}
.sidebar li {
    word-wrap: break-word; /* prevent words from continuing after sidebar */
}
.sidebar ul {
    width: 100%;
    list-style-type: none;
}
.sidebar li a {
    display: block;
    padding: 0.5em 0.5em 0.5em 1em;
}
.sidebar li.toctree-l2>a:before {
    content: '\f196'; /* arrow in box \f152';*/
    font-family: 'FontAwesome';
    margin: 0 0.5em 0 0;
    color: #777;
}
.sidebar li.toctree-l2.current>a:before {
    color: white;
}
.sidebar li.toctree-l2.show:not(.no-expand)>a:before{
    content: '\f147'; /* arrow down f150 */
    font-family: 'FontAwesome';
    margin: 0 0.5em 0 0;
    color: white;
}
.sidebar li:only-child a:before {
	content: "";
}
.sidebar li.toctree-l1 a:hover {
    color: #e6e6e6;
    background-color: #afafaf;
}
.sidebar li.toctree-l1>a {
    color: lightgray;
    font-size: 1.1em;
    font-weight: bold;
    padding: 0.4em 0.4em 0.4em 1.5em;
}
.sidebar li.toctree-l1.current>a, .sidebar li.toctree-l1.show>a {
    color: #404040;
    position: relative;
    background: #fcfcfc;
    border: none;
}
.sidebar li.toctree-l2>a:hover {
    color: #e6e6e6;
}
.sidebar li.toctree-l2>a {
    color: #4e4e4e;
    background-color: lightgray;
    font-size: 0.9em;
    font-weight: bold;
    padding: 0.4em 0.4em 0.4em 2em;
    list-style-type: circle;
}
.sidebar li.toctree-l2.current:not(.no-expand)>a, .sidebar li.toctree-l2.show:not(.no-expand)>a {
    background-color: #848484;
    color: #e6e6e6;
}
.sidebar li.toctree-l2>ul {
    display: none;
}
.sidebar li.toctree-l2.show>ul {
    display: block!important;
}
.sidebar li.toctree-l3 a {
    color: #4e4e4e;
    background-color: #afafaf;
    font-size: 0.8em;
    font-weight: bold;
    padding: 0.4em 0.4em 0.4em 5em;
    list-style-type: circle;
}
.sidebar li.toctree-l3 code {
    display: none;
}
.sidebar li.toctree-l3 a.current, .sidebar li.toctree-l3 a.show {
    color: black;
}



/* ************
 * LANDING PAGE
 * ************ */
.tensorly-functionalities-container {
    margin-left: 5em;
    margin-top: 2em;
    margin-bottom: 2em;
    background: inherit;
}
.tensorly-functionalities {
    text-align: left;
    font-size: 1.3em;
    background: inherit;
}
.hero-cta {
    padding-top: 3em;
    padding-bottom: 3em;
    background: #efefef;
}
.home-main {
    background: rgba(0, 209, 178, 0.47);
    padding-top: 8em;
    padding-bottom: 8em;
}
.home-hero {
    background: #e1feff;
}
a.box:hover {
    box-shadow: 0 4px 5px rgba(10, 10, 10, 0.62), 0 0 0 5px #66ffe8;
    background: white;
}
.tile.is-child {
    /* background: #efefef; */
	background: #f7f9f8;
	box-shadow: 0 4px 5px rgba(10,10,10,0.62), 0 0 0 5px #f2f9f8;
}



/*  *******
 * TOOLTIPS
 * ******** */
/* Tooltip text */
.tooltip .tooltiptext {
    visibility: hidden;
    background-color: #7a7a7a;
    color: #ffffff;
    text-align: center;
    border-radius: 6px;
    padding-left: 0.5em;
    padding-right: 0.5em;

    /* Position the tooltip text */
    position: absolute;
    z-index: 1;
    top: 75%;
	left: -150%;

    /* Fade in tooltip */
    opacity: 0;
    transition: opacity 1s;
}
/* Tooltip arrow
.tooltip .tooltiptext::after {
    content: "";
    position: absolute;
    bottom: 100%;
    left: 47%;
    border-width: 5px;
    border-style: solid;
    border-color: transparent transparent #555 transparent;
} */
/* Show the tooltip text when you mouse over the tooltip container */
.tooltip:hover .tooltiptext {
    visibility: visible;
    opacity: 0.8;
}


/* ******************
 * CODE RELATED STUFF
 * ****************** */

/* name of the functions, etc */
.docutils.literal {
    color: #27a290;
    background-color: whitesmoke;
}
dl.class > dd {
    padding-top: 2em;
    padding-bottom: 2em;
}
.descclassname {
    display: none;
}
.viewcode-link {
    float: right;
}
dl.function>dd, dl.class>dd, dl.method>dd {
	margin-left: 0.1em;
}
dl.function, dl.class, dl.method {
    border: none;
}
/* Methods */
dl.method>dt, dl.function>dt, dl.class>dt {
    padding: 0.5em;
    background-color: lightgray;
    font-size: 0.9em;
    display: block;
    margin: 0;
    background: #e8f4ff;
    font-family: monospace;
}
.function dd{
	margin-left: 0;
}
em.property {
    font-style: normal;
    color: #00b19b;
    font-weight: bold;
    font-size: 1.1em;
}
code {
    color: #27a293;
    font-family: monospace;
}
code.descname {
    background: transparent;
    font-size: 1.2em;
    font-weight: bold;
    color: #007365;
}
.admonition p {
	margin: 0!important;
}
.rubric, .admonition-title {
    padding: 0.5em 0.75em;
    font-weight: bold;
    color: #363636;
    border-top: 1px solid #dbdbdb;
    margin-top: 1em;
}
div.admonition dt {
    margin-left: 2em;
}
div.admonition dd {
    margin-left: 4em;
    margin-bottom: 1em;
}
/* Warning directive */
.admonition.warning {
    border: solid #9a2c42;
    margin: 1em 0;
    padding: 0;
    background: #e2d7d9;
}
.admonition.warning p{
    padding: 1em;
}
.admonition.warning p.first.admonition-title {
    padding: .5em .75em;
    margin: 0;
    font-weight: 700;
    color: #f5f5f5;
    background: #9a2c42;
    border-top: none;
    margin-top: 0;
}
/* Important directive */
.admonition.important {
    border: solid #356392;
    margin: 1em 0;
    padding: 0;
    background: #d2e5f9;
}
.admonition.important p{
    padding: 1em;
}
.admonition.important p.first.admonition-title {
    padding: .5em .75em;
    margin: 0;
    font-weight: 700;
    color: #f5f5f5;
    background: #467cb3;
    border-top: none;
    margin-top: 0;
}
/* Note directive */
.admonition.note {
    border: solid #008c77;
    padding: 0;
    margin: 1em 0;
    background: #b3f9ee;
}
.admonition.note p{
    padding: 1em;
}
.admonition.note p.first.admonition-title {
    padding: .5em .75em;
    margin: 0;
    font-weight: 700;
    color: #f5f5f5;
    background: #008c77;
    border-top: none;
    margin-top: 0;
}

/* ********
 * EXAMPLES 
 * ******** */
p.sphx-glr-signature>a {
    float: right;
    font-size: small;
    padding: 1em;
}
/* Content at the top of page */
#gallery-of-examples #contents {
    margin-bottom: 2em;
    padding: 1em;
    background: #f9f9f9;
}
div.sphx-glr-download a span:last-child {
    font-size: smaller;
}
.sphx-glr-thumbcontainer a {
    display: inline-block;
    padding: 10em 0.5em 2em 0.5em;
}
.sphx-glr-thumbcontainer {
    float: left;
    border: solid rgba(2, 218, 186, 0.18); float: left; margin: 0.75em; 
    height: 15em;
    width: 12em;
    border-width: 0.25em;
}
.sphx-glr-thumbcontainer p {
    text-align: center;
    margin-top: -11em;
    padding: 0;
}
.sphx-glr-thumbcontainer img {
    max-height: 100%;
    max-width: 100%;
    margin-bottom: 1em;
    margin-top: 1.5em;
}
.sphx-glr-thumbcontainer:hover {
    box-shadow: 0 0 15px #00f3cf;
}


/* download all buttons */
#gallery-of-examples table{
    display: none;
}
div.sphx-glr-footer {
    text-align: center;
    margin-top: 2em;
    margin-bottom: 1em;
	display: inline;
}
div.sphx-glr-download {
    display: inline;
    margin: 1em;
    vertical-align: middle;
}
div.sphx-glr-download a {
    background-color: #ccfff1;
    background-image: linear-gradient(to bottom, #c7fffc, rgba(0, 211, 204, 0.74));
    border-radius: 5px;
    border: 1px solid #00a086;
    color: #000;
    display: inline-block;
    display: table-caption;
    font-weight: bold;
    min-width: 20em;
    padding: 0.5em;
    text-align: center;
    font-size: 1em;
}
div.sphx-glr-download code.download::before {
    font-family: FontAwesome;
    content: "\f019\00A0";
}
div.sphx-glr-download code.download {
    display: inline-block;
    white-space: normal;
    word-break: normal;
    color: black;
    overflow-wrap: break-word;
    border: none;
    background: none;
}
/* generated by sphinx */
p.sphx-glr-signature>a {
    float: right;
    font-size: small;
    padding: 1em;
}
p.sphx-glr-signature {
	display: flow-root;
}
/* Content at the top of page */
#gallery-of-examples #contents {
    margin-bottom: 2em;
    padding: 1em;
    background: #f9f9f9;
}
div.sphx-glr-download a span:last-child {
    font-size: smaller;
}
.sphx-glr-thumbcontainer a {
    display: inline-block;
    padding: 10em 0.5em 2em 0.5em;
}
.sphx-glr-thumbcontainer {
    float: left;
    border: solid rgba(2, 218, 186, 0.18); float: left; margin: 0.75em; 
    height: 15em;
    width: 12em;
    border-width: 0.25em;
}
.sphx-glr-thumbcontainer p {
    text-align: center;
    margin-top: -11em;
    padding: 0;
}
.sphx-glr-thumbcontainer img {
    max-height: 100%;
    max-width: 100%;
    margin-bottom: 1em;
    margin-top: 1.5em;
}
.sphx-glr-thumbcontainer:hover {
    box-shadow: 0 0 15px #00f3cf;
}



