"""
The :mod:`tensorly.contrib.decomposition` module includes utilities for performing
tensor decomposition for which the API is still evolving.                                                                                               
"""

from .mps_decomposition_cross import matrix_product_state_cross

