"""                                                                                                                                                      
The :mod:`tensorly.regression` module includes classes for performing Tensor
Regression.
"""

from .kruskal_regression import KruskalRegressor
from .tucker_regression import TuckerRegressor
